


AE_Utilities = {}

function AE_Utilities:Version()
	return "1.0"
end

function AE_Utilities:GetBezierValue(curve, pointNum, frame, prePoint, getModifiedPoints)
    local curvature = curve:GetCurvature(pointNum, frame)
    local weight = curve:GetWeight(pointNum, frame, prePoint)
    local offset = curve:GetOffset(pointNum, frame, prePoint)
    local pointToRange = function(p)
      if curve.fClosed then
         if p < 0 then p = curve:CountPoints()-1 end
         if p >= curve:CountPoints() then p = 0  end
      else
         if p < 0 then p = 0 end
         if p >= curve:CountPoints() then p = curve:CountPoints()-1 end    
      end
      return p    
    end
    local postPointNum = pointToRange(pointNum + 1)
    local prePointNum = pointToRange(pointNum - 1)
    local secondPointNum = postPointNum
    if prePoint then secondPointNum = prePointNum end
    if secondPointNum == pointNum then 
      if getModifiedPoints then 
        return curve:Point(pointNum).fPos
      else 
        return curve:Point(pointNum).fAnimPos:GetValue(frame)
      end
    end
    local N = curve:Point(postPointNum).fAnimPos:GetValue(frame) - curve:Point(prePointNum).fAnimPos:GetValue(frame)
    if getModifiedPoints then N = curve:Point(postPointNum).fPos - curve:Point(prePointNum).fPos end
    N:NormMe()
    if prePoint then N = N*-1 end
    local L = (curve:Point(secondPointNum).fAnimPos:GetValue(frame) - curve:Point(pointNum).fAnimPos:GetValue(frame)):Mag()
    if getModifiedPoints then L = (curve:Point(secondPointNum).fPos - curve:Point(pointNum).fPos):Mag() end
    L = L * curvature * weight
    local vec = N * L
    vec:Rotate(offset)
    if getModifiedPoints then 
      return curve:Point(pointNum).fPos + vec
    else
      return curve:Point(pointNum).fAnimPos:GetValue(frame) + vec
    end
end

function AE_Utilities:GetAngleBetween(vec1, vec2)
  	v1 = vec1:Norm()
		v2 = vec2:Norm()
		local angle = v1:Dot(v2)
		if (angle > 1.0) then
			angle = 0
		elseif (angle < -1.0) then
			angle = math.pi
		else
			angle = math.acos(angle)
		end
		local norm = v1:Cross(v2)
		if (norm.z < 0.0) then
			angle = -angle
		end
    return angle
end

function AE_Utilities:SumActionInfluences(moho, frame, derivedChannel, layer)
--print("starting to sum influences")
   local sum = 0
   local defaultValue = derivedChannel:GetValue(0)
   local parentSceleton = self:FindControllingSkeleton(moho, layer)
   if not parentSceleton then return 0 end
   
   for actID = 0, derivedChannel:CountActions()-1 do
       local actName = derivedChannel:ActionName(actID)
       local actChannel = moho:ChannelAsAnimVal(derivedChannel:Action(actID))
       local boneName = actName
       if string.sub(boneName, string.len(boneName)-1) == " 2" then boneName = string.sub(boneName, 1, string.len(boneName)-2) end
       local smartBone = parentSceleton:BoneByName(boneName)
       if smartBone and actName ~= layer:CurrentAction() then
          local boneTrack = smartBone.fAnimAngle:ActionByName(actName)
          if boneTrack then
            local curSmartboneValue = smartBone.fAnimAngle:GetValue(frame) -- bad for nested smartbones
			--print(smartBone:Name(), ": ", curSmartboneValue*180/math.pi, " ", smartBone.fAngle*180/math.pi)
			curSmartboneValue = smartBone.fAngle -- better for nested smartbones
            local smartBoneFrame, smartBoneSubFrame = self:FindSmartBoneFrame(curSmartboneValue, moho:ChannelAsAnimVal(boneTrack))
            if(smartBoneFrame>0 or smartBoneSubFrame ~= smartBoneFrame) then
              local nextVal = actChannel:GetValue(smartBoneFrame) - defaultValue
			  if smartBoneSubFrame ~= smartBoneFrame then
				nextVal = (actChannel:GetValue(1) - defaultValue)*smartBoneSubFrame  
			  end
              sum = sum + nextVal 
            end      
          end
       end
   end
   return sum
end

function AE_Utilities:FindSmartBoneFrame(angle, smartBoneTrack)

  local foundFrame = -1
  local foundSubFrame = -1
  local defAngle = smartBoneTrack:GetValue(0)
  if defAngle == angle then 
    return 0,0
  end
  for f=1, smartBoneTrack:Duration() do
    local curVal = smartBoneTrack:GetValue(f)
    local preVal = smartBoneTrack:GetValue(f-1)
    if ((curVal-angle)*(preVal-angle)) <= 0 then
       foundFrame = f-1
       foundSubFrame = f-1+(angle-preVal)/(curVal-preVal)
       --if math.abs(preVal-angle) < math.abs(curVal - angle) then
       --   foundFrame = f-1
       --end
    end
    if f==smartBoneTrack:Duration() and foundFrame == -1 and math.abs(curVal-angle)<math.abs(preVal-angle) then 
      foundFrame = f 
      foundSubFrame = f
    end 
  end
  if smartBoneTrack:Duration() == 1 then return foundFrame, foundSubFrame end
  return foundFrame, foundFrame
end

function AE_Utilities:FindControllingSkeleton(moho, layer)
	local boneLayer = nil	
	if layer:LayerType() == MOHO.LT_BONE then
		boneLayer = moho:LayerAsBone(layer)
		return boneLayer:Skeleton(), boneLayer
	end
	local parent = layer

	while parent:Parent() do
		parent = parent:Parent()
		if parent:LayerType() == MOHO.LT_BONE then
			local boneLayer = moho:LayerAsBone(parent)
			if boneLayer and boneLayer:Skeleton():CountBones()>0 then
				return boneLayer:Skeleton(), boneLayer
			end
		end
	end
	if boneLayer then return boneLayer:Skeleton(), boneLayer end
	return nil
end

